#include "main.h"
#include "monitoring.h"
#include "com/debug.h"

#include "type_coerce.h"
#include "payload_parser.h"

#include "vec_match.h"


#define data_page_addr 0x0801EC00UL

static volatile const float *stored_vector = (volatile const float *)(data_page_addr + 1024);
static volatile const reference_info* stored_info = (volatile const reference_info *)(data_page_addr);

/** process */
void rpt_process_captured_vector(reference_info *nfo)
{
	const int count = HEALTH_REF_SAMP_LEN;

	samples_to_milliamps(samp_buf.floats, samp_buf.uints, count);

	// Calculate whatever we want to calculate from the waveform
	arm_rms_f32(samp_buf.floats, count, &nfo->rms);

	// convert to a FFT vector; in place.
	samples_to_fft(samp_buf.floats, samp_buf.floats, count);

	nfo->magic = REF_MAGIC;
}


bool rpt_is_ref_stored(void)
{
	return stored_info->magic == REF_MAGIC;
}


float rpt_calc_vector_dist(void)
{
	VecHausdorffCfg cfg;
	cfg.length = HEALTH_REF_FFT_LEN;

	// this needs fine tuning
	cfg.cost_x = RPT_COST_X;
	cfg.cost_y = RPT_COST_Y;
	cfg.max_drift_x = 5;

	return vec_hausdorff(samp_buf.floats, (const float *)stored_vector, &cfg);
}


/** store reference vector to flash */
bool rpt_store_reference(void)
{
	reference_info nfo;
	rpt_process_captured_vector(&nfo);

	FLASH_Status status;
	FLASH_Unlock();

	// --- map ---
	// INFO PAGE
	// 0-127
	// 128-255
	// 256-383
	// 384-511

	// pages have 1024 bytes

	info("Erasing data flash.");
	for (uint8_t i = 0; i < 5; i++) {
		uint32_t addr = data_page_addr + i*1024;
		dbg("Erasing page at 0x%08"PRIx32, addr);
		status = FLASH_ErasePage(addr);
		if (status != FLASH_COMPLETE) goto fail;
	}

	info("Writing reference data.");
	uint32_t addr = data_page_addr + 1024; // start of data pages
	for (uint32_t i = 0; i < HEALTH_REF_FFT_LEN; i++, addr+=4) {
		if ((i & 0x3F) == 0) dbg("Write bin[%"PRIu32"] @ 0x%08"PRIx32, i, addr); // print only some...
		status = FLASH_ProgramWord(addr, ((union conv32){.f32 = samp_buf.floats[i]}).u32);
		if (status != FLASH_COMPLETE) goto fail;
	}

	// finally store the data page
	void *nfo_b = (void*)&nfo;
	addr = data_page_addr;
	PayloadParser pp = pp_start(nfo_b, sizeof(reference_info));

	info("Writing info block.");
	for(size_t i = 0; i < sizeof(reference_info); i+=4, addr+=4) { // step by words
		uint32_t word = pp_u32(&pp);
		status = FLASH_ProgramWord(addr, word);
		if (status != FLASH_COMPLETE) goto fail;
	}

	FLASH_Lock();

	info("Store success!");
	return true;

fail:
	error("Flash operation failed.");

	// try to nuke the data page, just in case
	FLASH_ErasePage(data_page_addr);

	FLASH_Lock();
	return false;
}
