#include "wifistatusled.h"
#include "utils/timebase.h"
#include "gui/colorled.h"
#include "com/debug.h"

// public field
struct wifistatus WifiStatus = {};

#define CLED_WIFI 0
#define CLED_ACQ 1

#define WIFILED_LIGHT_MS 300
#define WIFILED_DELAY_MS 150

#define FAST_BLINK_TIME 2000
#define SLOW_BLINK_TIME 4000

// extern
static bool blinkmode_fast = true;

static task_pid_t blinktask_slow;
static task_pid_t blinktask_fast;

static bool wps_was_active = false;

static uint32_t leds[2] = {
	WIFILED_IDLE_COLOR,
	ACQLED_IDLE_COLOR
};


static void color_status_show(void)
{
	colorled_set_many(leds, 2);
}


void set_acqstatus_led(uint32_t color)
{
	leds[CLED_ACQ] = color;
	color_status_show();
}


static void cled_off(void *which)
{
	leds[(uint32_t)which] = (which == CLED_WIFI ? WIFILED_IDLE_COLOR : ACQLED_IDLE_COLOR);
	color_status_show();
}


void blink_wifistatus_led(uint32_t color, ms_time_t ms)
{
	leds[CLED_WIFI] = color;
	color_status_show();

	schedule_task(cled_off, (void*)CLED_WIFI, ms, false);
}


void blink_acqstatus_led(uint32_t color, ms_time_t ms)
{
	leds[CLED_ACQ] = color;
	color_status_show();

	schedule_task(cled_off, (void*)CLED_ACQ, ms, false);
}


#define WPS_COLOR rgb(128,104,0)
#define WPS_BUSY_COLOR rgb(128,30,0)

#define AP_COLOR  rgb(0,40,128)
#define STA_COLOR_CON  rgb(16,128,0)
#define STA_COLOR_DIS  rgb(128,0,0)


static void show_sta_status(void *unused)
{
	(void)unused;

	uint32_t color = WifiStatus.sta_connected ? STA_COLOR_CON : STA_COLOR_DIS;
	blink_wifistatus_led(color, WIFILED_LIGHT_MS);
}

/** set blink speed */
static void set_blinkspeed_fast(bool fast)
{
	if (fast) {
		reset_periodic_task(blinktask_fast);
	} else {
		reset_periodic_task(blinktask_slow);
	}

	enable_periodic_task(blinktask_fast, fast);
	enable_periodic_task(blinktask_slow, !fast);

	blinkmode_fast = fast;
}

/** cb for the following func */
static void leave_fastblink_mode(void *unused)
{
	(void)unused;
	set_blinkspeed_fast(false);
}

/** slow down w/ a delay */
static void delayed_blinkslow(void)
{
	blinkmode_fast = false; // set in advance
	schedule_task(leave_fastblink_mode, NULL, (FAST_BLINK_TIME + 100), false);
}

void wifistatus_onchange(void)
{
	if (WifiStatus.wps_active) {
		WifiStatus.wps_preparing = false;
		leds[CLED_WIFI] = WPS_COLOR;
		wps_was_active = true;

	} else if (WifiStatus.wps_preparing) {
		leds[CLED_WIFI] = WPS_BUSY_COLOR;

	} else if (wps_was_active) {
		wps_was_active = false;
		leds[CLED_WIFI] = WIFILED_IDLE_COLOR;
	}

	color_status_show();

	if (WifiStatus.opmode != OPMODE_AP) {
		if (!WifiStatus.sta_connected) {

			if (!blinkmode_fast)
				set_blinkspeed_fast(true); // alert!

		} else {
			if (blinkmode_fast)
				delayed_blinkslow(); // slow down w/ a delay
		}
	} else {
		// ap only - slow blink
		if (blinkmode_fast)
			delayed_blinkslow();
	}

//	if (!WifiStatus.reporting_enabled) {
//		leds[CLED_ACQ] = 0; // turn rpt led off
//	}
}


static void periodic_show_wifi_status(void *unused)
{
	(void)unused;

	if (!WifiStatus.wps_active && !WifiStatus.wps_preparing) {
		switch (WifiStatus.opmode) {
			case OPMODE_STA_AP:
				blink_wifistatus_led(AP_COLOR, WIFILED_LIGHT_MS);
				schedule_task(show_sta_status, NULL, WIFILED_LIGHT_MS + WIFILED_DELAY_MS, false);
				break;

			case OPMODE_STA:
				show_sta_status(NULL);
				break;

			case OPMODE_AP:
				blink_wifistatus_led(AP_COLOR, WIFILED_LIGHT_MS);
				break;

			default: // fail
				blink_wifistatus_led(FAIL_COLOR, 1000);
		}
	}
}


void wifistatusled_init(void)
{
	blinktask_slow = add_periodic_task(periodic_show_wifi_status, NULL, SLOW_BLINK_TIME, false);
	blinktask_fast = add_periodic_task(periodic_show_wifi_status, NULL, FAST_BLINK_TIME, false);

	set_blinkspeed_fast(true);
}
