#include "main.h"

#include "hw_utils/all.h"
#include "hw_init.h"

#include "utils/timebase.h"
#include "utils/str_utils.h"
#include "utils/circbuf.h"
#include "utils/mutex.h"

#include "gui/infoled.h"
#include "gui/colorled.h"
#include "gui/buttons.h"
#include "gui/wifistatusled.h"

#include "com/com_fileio.h"
#include "com/debug.h"
#include "com/datalink.h"
#include "com/sbmp_control.h"

#include "bus/event_queue.h"
#include "bus/event_handler.h"

#include "acquire/sampling.h"
#include "acquire/acquire_control.h"

#include "sbmp.h"
#include "globals.h"

// ----


/** Called continuously in the main loop */
void poll_subsystems(void)
{
	// poll serial buffers (runs callback)
	com_poll(debug_iface);
	com_poll(data_iface);

	// run queued tasks
	tq_poll();

	// handle queued events
	Event evt;
	until_timeout(2) { // take 2 ms max
		if (eq_take(&evt)) {
			run_event_handler(&evt);
		} else {
			break;
		}
	}
}

/** This is only for debugging, not really useful */
static void byte_received_on_debug_iface(ComIface *iface)
{
	uint8_t b;
	while (com_rx(iface, &b)) {
		dbg("dbg rx %d", b);
	}
}

/** Attempt a handshake on SBMP - if hsk not OK yet. */
void try_handshake(void)
{
	if (sbmp_ep_handshake_status(dlnk_ep) != SBMP_HSK_SUCCESS) {
		sbmp_ep_start_handshake(dlnk_ep);
	}
}


void btn_press(void)
{
	info("--- BTN!! ---");

	//wificontrol_set_ap();

	wificontrol_start_wps();
}


int main(void)
{
	hw_init();

	uint8_t rst_reason = get_reset_reason_raw();

	if (rst_reason) esp_reset(true);

	printf(DEBUG_EOL);
	banner("*** AC current analyser - acquisition module ***");
	printf(DEBUG_EOL);
	banner_info("(c) Ondrej Hruska, 2016");
	banner_info("Katedra mereni FEL CVUT");
	banner_info("");
	banner_info("Version "FIRMWARE_VERSION", built "__DATE__" at "__TIME__);
	banner_info("");
	banner_info("Reset reason: %s (%02xh)", get_reset_reason_str(), rst_reason);
	printf(DEBUG_EOL);

	clear_reset_reason();

	// ---

	// 3 white blinks
	for (int i = 0; i < 3; i++) {
		blink_wifistatus_led(rgb(128, 128, 150), 50);
		delay_ms(200);
	}

	/* Indicate reset reaosn by a coloured blink */

	uint32_t color;

	switch (rst_reason) {
		case RST_PIN:
			// reset button pressed, also reset ESP
			color = rgb(0, 128, 0);
			break;

		case RST_POR:
		case RST_START:
			color = rgb(0, 128, 0);
			break;

		default:
			color = rgb(128, 0, 0); // red
			break;
	}

	blink_wifistatus_led(color, 500);
	delay_ms(500);

	debug_iface->rx_callback = byte_received_on_debug_iface;

	btn_set_press_callback(BTN1, btn_press);


	// Periodic SBMP attempt
	add_periodic_task(try_handshake, 700, true); // ESP uses 1000ms. The time difference avoids a possible conflict.

	add_periodic_task(SoftWDG_Tick1ms, 1, false); // no queue.

	// ---

	esp_reset(false); // deassert ESP reset

	ms_time_t t = ms_now();

	while (1) {
		poll_subsystems();

		// feed the Idle dog
		if (!acq_dma_in_progress() && !acq_is_running()) {
			SoftWDG_Feed(SoftWDG_ACQ_IDLE);
		}

		// indicate correct program operation by flashing the green LED.
		if (ms_loop_elapsed(&t, 500)) {
			led_toggle(LED_G);

			// Reset the watchdog
			// This also ensures that the timebase is running well
			WDG_Feed();
		}
	}
}





/*  // here's example of writing flash
	// erase data page

	#define page_addr 0x0803FF00
	#define var_addr 0x0803FF00

	FLASH_Status status;
	FLASH_Unlock();
	status = FLASH_ErasePage(page_addr);
	dbg("erase status = %d", status);
	status = FLASH_ProgramWord(var_addr, 0x0F5500FF);
	dbg("program status = %d", status);
	FLASH_Lock();

	dbg("Read = %08"PRIx32, *(volatile uint32_t *)(var_addr));



	FLASH_Unlock();
	status = FLASH_ErasePage(page_addr);
	dbg("erase status = %d", status);
	status = FLASH_ProgramWord(var_addr, 0xAA00FF00);
	dbg("program status = %d", status);
	FLASH_Lock();

	dbg("Read = %08"PRIx32, *(volatile uint32_t *)(var_addr));


*/




/* // EXAMPLE OF CHAINED EVENT HANDLERS
 *  uint32_t h1 = register_event_handler(EVT_ACQ_DMA_COMPLETE, acquire_task_done);
	uint32_t h2 = register_event_handler(EVT_ACQ_DMA_FAILED, acquire_task_done);

	if (!h1 || !h2) {
		error("Failed to register handlers");
		// remove both, 0 -> no-op.
		remove_event_handler(h1);
		remove_event_handler(h2);
	} else {
		chain_event_handler(h1, h2, false); // reciprocal chain.
		dbg("Handlers are now chained.");

		dbg("exist? %d %d", event_handler_exists(h1), event_handler_exists(h2));

		remove_event_handler(h2);

		dbg("exist? %d %d", event_handler_exists(h1), event_handler_exists(h2));
	}
	*/
